
from collections import abc
from functools import partial
import inspect
import re
import typing as t
from collections.abc import Callable, Hashable
from copy import deepcopy
from importlib import import_module
from typing import ForwardRef, TypeVar

from typing_extensions import Self

_object_setattr = object.__setattr__
_setattr = setattr

_T = TypeVar('_T')


def ordered_set(it: abc.Iterable[_T]) -> abc.Set[_T]:
    return dict.fromkeys(it).keys()


def private_setattr(klass=None, *, name: str='setattr', setattr=True, setattr_fn=_object_setattr, frozen: str =None):

    def decorator(cls_):

        # def _setfunc(self: Self, force=False):
        #     if not force and frozen and getattr(self, frozen, False):
        #         return _setattr
        #     else:
        #         return setattr_fn

        def setter(self: Self, name=None, value=None, force=False, /, **kw):
            if not force and frozen and getattr(self, frozen, False):
                setter_ = _setattr
            else:
                setter_ = setattr_fn
            
            name and kw.setdefault(name, value)
            for k,v in kw.items():
                setter_(self, k, v)
        
        def __setattr__(self: Self, name: str, value):
            if self._privateattr_regex.search(name):
                # if not (frozen or getattr(self, frozen, False)):
                return setattr_fn(self, name, value)
            getattr(self, name)
            raise AttributeError(f"`cannot set {name!r} on frozen {self.__class__.__qualname__!r}.")

        _base__init_subclass__ = cls_.__init_subclass__
        
        def __init_subclass__(cls: type[Self], **kwargs):
            if not hasattr(cls, fn := f'_{cls.__name__.lstrip("_")}__{name}'):
                _setattr(cls, fn, setter)
            pre = r'|'.join(f'_{c.__name__.lstrip("_")}__' for c in cls.__mro__)
            cls._privateattr_regex = re.compile(f"^(?:{pre}).+")
            _base__init_subclass__(**kwargs)
        cls_.__init_subclass__ = classmethod(__init_subclass__)

        pre = r'|'.join(f'_{c.__name__.lstrip("_")}__' for c in cls_.__mro__)
        cls_._privateattr_regex = re.compile(f"^(?:{pre}).+")

        if not hasattr(cls_, fn := f'_{cls_.__name__.lstrip("_")}__{name}'):
            _setattr(cls_, fn, setter)

        if setattr and cls_.__setattr__ is _object_setattr:
            cls_.__setattr__ = __setattr__
        
        return cls_

    return decorator if klass is None else decorator(klass)




def typed_signature(
    callable: Callable[..., t.Any], *, follow_wrapped=True, globalns=None, localns=None
) -> inspect.Signature:
    sig = inspect.signature(callable, follow_wrapped=follow_wrapped)

    if follow_wrapped:
        callable = inspect.unwrap(
            callable, stop=(lambda f: hasattr(f, "__signature__"))
        )

    if globalns is None:
        if not (globalns := getattr(callable, "__globals__", None)): 
            if isinstance(callable, partial):
                callable = callable.func
            getattr(
            import_module(callable.__module__), "__dict__", None
        )

    params = (
        p.replace(annotation=eval_type(p.annotation, globalns, localns))
        for p in sig.parameters.values()
    )

    return sig.replace(
        parameters=params,
        return_annotation=eval_type(sig.return_annotation, globalns, localns),
    )


def eval_type(value, globalns, localns=None):

    if isinstance(value, str):
        value = ForwardRef(value)
    try:
        return t._eval_type(value, globalns, localns)
    except NameError: # pragma: no cover
        return value




class MissingType:

    __slots__ = ()

    __value__: t.ClassVar['MissingType'] = None

    def __new__(cls):
        return cls.__value__

    @classmethod
    def _makenew__(cls, name):
        if cls.__value__ is None:
            cls.__value__ = object.__new__(cls)
        return cls()

    def __bool__(self): return False

    def __str__(self): return ''

    def __repr__(self): return f'Missing'

    def __reduce__(self):
        return self.__class__, () # pragma: no cover
    
    def __eq__(self, x):
        return x is self

    def __hash__(self):
        return id(self)







Missing = MissingType._makenew__('Missing')






_T_Key = t.TypeVar("_T_Key")
_T_Val = t.TypeVar("_T_Val", covariant=True)
_T_Default = t.TypeVar("_T_Default", covariant=True)


class ReadonlyDict(dict[_T_Key, _T_Val]):
    """A readonly `dict` subclass.

    Raises:
        TypeError: on any attempted modification
    """

    __slots__ = ()


    def not_mutable(self, *a, **kw):
        raise TypeError(f"readonly type: {self} ")

    __delitem__ = __setitem__ = setdefault = not_mutable
    clear = pop = popitem = update = __ior__ = not_mutable
    del not_mutable

    @classmethod
    def fromkeys(cls, it: abc.Iterable[_T_Key], value: _T_Val=None):
        return cls((k, value) for k in it)

    def __reduce__(self):
        return (
            self.__class__,
            (dict(self),),
        )

    def copy(self):
        return self.__class__(self)
    __copy__ = copy

    def __deepcopy__(self, memo=None):
        return self.__class__(deepcopy(dict(self), memo))

    __or = dict[_T_Key, _T_Val].__or__

    def __or__(self, o):
        return self.__class__(self.__or(o))
        


class FrozenDict(ReadonlyDict[_T_Key, _T_Val]):
    """An hashable `ReadonlyDict`
    """
    __slots__ = "_v_hash",

    def __hash__(self):
        try:
            ash = self._v_hash
        except AttributeError:
            ash = None
            items = self._eval_hashable()
            if items is not None:
                try:
                    ash = hash(items)
                except TypeError:
                    pass
            _object_setattr(self, '_v_hash', ash)

        if ash is None:
            raise TypeError(f"un-hashable type: {self.__class__.__name__!r}")

        return ash

    def _eval_hashable(self) -> Hashable:
        return *((k, self[k]) for k in sorted(self)),




from abc import ABC, ABCMeta, abstractmethod
from itertools import groupby, zip_longest
import re
import sys
from types import MappingProxyType
import typing as t
from logging import getLogger
from collections import ChainMap, abc, defaultdict
from typing_extensions import Self
from weakref import WeakKeyDictionary, WeakSet, WeakValueDictionary, ref
from importlib import import_module
from .exceptions import ProError


from . import signals
from .markers import GUARDED, PRIVATE, PROTECTED, PUBLIC, AccessLevel, _PredicateOpsMixin, Injectable, ProInvertPredicate, ProPredicate, is_injectable
from ._common import ReadonlyDict, ordered_set, private_setattr, FrozenDict
from .providers import Provider, ProviderRegistryMixin


if t.TYPE_CHECKING: # pragma: no cover
    from .graph import DepGraph, DepKey, DepSrc


logger = getLogger(__name__)

_dict_setdefault = dict.setdefault
_dict_setitem = dict.__setitem__
_object_new = object.__new__


def _calling_module(depth=2) -> t.Optional[str]:
    """Get the globals() or locals() scope of the calling scope"""
    name = sys._getframe(depth).f_globals.get('__name__')
    try:
        name and import_module(name) 
    except Exception: # pragma: no cover
        return
    else:
        if name == __name__:
            name = _calling_module(depth+2)    
        return name



class _ContainerRegistry(ReadonlyDict[str, dict['BaseContainer', None]]):
    __slots__ = ()

    __getitem = dict[str, dict['BaseContainer', None]].__getitem__
    __get = dict[str, dict['BaseContainer', None]].get
    __contains = dict[str, dict['BaseContainer', None]].__contains__

    _placeholders = {
        '**': r'.*', # '[^{placeholders}]*',
        '*': r'[^.|:]*', #'[^{placeholders}.]*',
        '++': r'.+', # '[^{placeholders}]+',
        '+': r'[^.|:]+', # '[^{placeholders}.]+',
    }

    _pattern_split_re = re.compile(f"({'|'.join(map(re.escape, _placeholders))})")
    pfmt = { 'placeholders': re.escape(''.join({*''.join(_placeholders)})) }
    for k in _placeholders:
        _placeholders[k] = _placeholders[k].format_map(pfmt)
    del k, pfmt
       
    @classmethod
    def _complie_pattern(cls, pattern: t.Union[str, re.Pattern]):
        if isinstance(pattern, re.Pattern):
            return pattern
        parts = cls._pattern_split_re.split(pattern)
        pattern = ''.join(cls._placeholders.get(x, re.escape(x)) for x in parts)
        return re.compile(f'^{pattern}$')

    def add(self, *instances: 'BaseContainer'):
        for inst in instances:
            inst._is_anonymous or self.__getitem(inst.qualname).setdefault(inst)
    
    def get(self, key: str, default = None):
        ret = self.__get(key)
        return default if ret is None else tuple(ret)

    @t.overload
    def all(self, 
            *patterns: t.Union[str, re.Pattern], 
            module: t.Union[str, list[str], tuple[str]]=None, 
            name: t.Union[str, list[str], tuple[str]]=None, 
            group: t.Literal[False]=False) -> abc.Generator['BaseContainer', None, None]: ...

    @t.overload
    def all(self, 
            *patterns: t.Union[str, re.Pattern], 
            module: t.Union[str, list[str], tuple[str]]=None, 
            name: t.Union[str, list[str], tuple[str]]=None, 
            group: t.Literal[True]=True) -> abc.Generator[tuple['BaseContainer'], None, None]: ...

    def all(self, 
            *patterns: t.Union[str, re.Pattern], 
            module: t.Union[str, list[str], tuple[str]]=None, 
            name: t.Union[str, list[str], tuple[str]]=None, 
            group: bool=False):
        
        if module or name:
            if not isinstance(module, (list, tuple)):
                module = f'{module or "**"}',
            if not isinstance(name, (list, tuple)):
                name = f'{name or "**"}',

            lm, ln = len(module), len(name)
            if lm == ln:
                it = zip(module, name)
            elif lm == 1:
                it = zip(module * ln, name)
            elif ln == 1:
                it = zip(module, name * lm)
            else:
                raise ValueError(
                    f'`module` and `name` lists can either be of equal lengths '
                    f'or one of them must contain a single item.'
                )
            patterns = tuple(':'.join(mn) for mn in it) + patterns
        elif not patterns:
            patterns = "**",

        seen = set()
        for pattern in (self._complie_pattern(p) for p in patterns):
            for k, v in self.items():
                if not (k in seen or not pattern.search(k) or seen.add(k)):
                    if group:
                        yield tuple(v)
                    else:
                        yield from v

    @t.overload                        
    def find(self, *patterns: t.Union[str, re.Pattern], module: str=None, name: str=None, group: bool=True): ...
    def find(self, *a, **kw):
        for v in self.all(*a, **kw):
            return v

    def __contains__(self, k: t.Union[str, 'BaseContainer']):
        if isinstance(k, BaseContainer):
            return k in self.__get(k.qualname, ())
        return self.__contains(k)

    def __getitem__(self, k: str) -> tuple['BaseContainer']:
        return tuple(self.__getitem(k))

    def __missing__(self, key: str):
        return _dict_setdefault(self, key, WeakKeyDictionary())

    def __repr__(self):
        return f'{self.__class__.__name__}({({k: self[k] for k in self})})'



class ProEntries(FrozenDict['BaseContainer', None]):
    
    __slots__ = ()

    __contains = dict.__contains__
    is_atomic: bool = False

    @classmethod
    def make(cls, it: abc.Iterable['BaseContainer']=()):
        return cls((v,None) for v in it)
    fromkeys = make

    def atomic(self):
        return AtomicProEntries.make(self)

    def _eval_hashable(self):
        return tuple(self.atomic())
    
    __hash__ = FrozenDict.__hash__

    def __contains__(self, k: 'BaseContainer') -> bool:
        if self.__contains(k):
            return True
        elif getattr(k, 'is_atomic', True):
            return not self.is_atomic and any(k in c for c in self if not c.is_atomic)
        else:
            return all(a in self for a in k.atomic) 

    def __eq__(self, o: Self) -> bool:
        if isinstance(o, ProEntries):
            return self._eval_hashable() == o._eval_hashable()
        elif isinstance(o, abc.Mapping):
            return False
        return NotImplemented

    def __ne__(self, o: Self) -> bool:
        if isinstance(o, ProEntries):
            return self._eval_hashable() != o._eval_hashable()
        elif isinstance(o, abc.Mapping):
            return True
        return NotImplemented


class AtomicProEntries(ProEntries):
    
    __slots__ = ()

    is_atomic: bool = True
    
    @classmethod
    def make(cls, it: abc.Iterable['Container']=()):
        return cls((a,None) for at in it for a in at.atomic)

    def atomic(self):
        return self

    def _eval_hashable(self):
        return tuple(self)
    


class ContainerMeta(ABCMeta):

    _registry: t.ClassVar[_ContainerRegistry] = _ContainerRegistry()
    register: t.Final = ABCMeta.register

    def __call__(self, *args, **kwds):
        if not 'module' in kwds:
            kwds['module'] = _calling_module()
        res: Container = super().__call__(*args, **kwds)
        self._registry.add(res)
        return res



@ProPredicate.register
@private_setattr(frozen='_pro')
class BaseContainer(_PredicateOpsMixin, metaclass=ContainerMeta):
    
    __slots__ = ()
    is_atomic: bool = True
    
    @classmethod
    @abstractmethod
    def _collect(cls, *a, **kw) -> 'Group': ...

    @property
    def _is_anonymous(self) -> bool:
        """`True` if this container can be added to the registry or `False` if 
            otherwise.
        """
        return not self.name
    
    @property
    @abstractmethod
    def atomic(self):
        """`AtomicProEntries`(s) 
        """
        
    @property
    @abstractmethod
    def bases(self) -> abc.Mapping[Injectable, Provider]:
        """the base container.
        """

    @property
    @abstractmethod
    def providers(self) -> abc.Mapping[Injectable, Provider]:
        """A mapping of providers registered in the container.
        """
        
    @property
    @abstractmethod
    def _pro(self) -> t.Optional[FrozenDict[Self, int]]:
        """The container's provider resolution order."""
        
    @property
    def pro(self) -> FrozenDict[Self, int]:
        """The container's provider resolution order.
        
        Like python's class `__mro__` the `pro` is computed using 
        [C3 linearization](https://en.wikipedia.org/wiki/C3_linearization)

        Returns:
            pro (FrozenDict[Container, int]): 
        """
        if not None is (pro := self._pro):
            return pro
        self.__setattr(_pro=self._evaluate_pro())
        return self._pro     
 
    @property
    def qualname(self) -> str:
        return f'{self.module}:{self.name}'

    @property
    @abstractmethod
    def g(self) -> ReadonlyDict['DepGraph', 'DepGraph']: ...

    @property
    @abstractmethod
    def module(self) -> str: ...

    @property
    @abstractmethod
    def name(self) -> str: ...

    def extends(self, other: Self) -> bool:
        """Check whether this container extends the given base. 
        
        Params:
            base (Container): The base container to check

        Returns:
            bool:
        """
        return other in self.pro
        # if other.is_atomic:
        #     return other in self.pro
        # else:
        #     return all(self.extends(x) for x in other.atomic)

    def get_graph(self, base: 'DepGraph'):
        try:
            return self.g[base]
        except KeyError:
            return _dict_setdefault(self.g, base, self.create_graph(base))

    def create_graph(self, base: 'DepGraph'):
        return DepGraph(self, base)


    def pro_entries(self, it: abc.Iterable['Container'], bindings: 'DepGraph', src: 'DepSrc') -> abc.Iterable['Container']:
        pro = self.pro
        return tuple(c for c in it if c in pro)
        
    def _evaluate_pro(self):
        if self.is_atomic:
            res, bases = {self:0}, [*self.bases]
        else:
            res, bases = {}, [*self.atomic]

        if bases:
            i, miss = 0, 0
            ml = [*([*b.pro] for b in bases), [*bases]]
            while ml:
                if i == len(ml):
                    if miss >= i:
                        raise ProError(f'Cannot create a consistent provider resolution order {miss=}, {ml=}')
                    i = 0
                ls = ml[i]
                h = ls[0]
                if h in res:
                    pass
                elif any(l.index(h) > 0 for l in  ml if not l is ls and h in l):
                    i += 1
                    miss += 1
                    continue
                else:
                    res[h] = i
                ls.pop(0)
                miss = 0
                if ls:
                    i += 1
                else:
                    ml.pop(i)

        return AtomicProEntries.fromkeys(res)

    def __eq__(self, o) -> bool:
        if isinstance(o, BaseContainer):
            return self is o
        return NotImplemented

    def __ne__(self, o) -> bool:
        if isinstance(o, BaseContainer):
            return not self is o
        return NotImplemented

    def __hash__(self) -> int:
        return id(self)

    def __or__(self, o):
        if isinstance(o, BaseContainer):
            return self._collect((self, o))
        else:
            return super().__or__(o)
    
    __ior__ = __or__

    # def __ror__(self, o):
    #     if isinstance(o, BaseContainer):
    #         return self._collect((o, self))
    #     else:
    #         return super().__ror__(o)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.qualname!r})"




class Container(BaseContainer, ProviderRegistryMixin):
    """A mapping of dependencies to their providers. We use them to bind 
    dependencies to their providers. 
   
    Attributes:
        name (str): The container's name
        bases (tuple[Container]): The container's bases
        default_access_level (AccessLevel): The default `access_level` to assign 
        to providers registered in this container
    """
    __slots__ = 'module', 'name', 'providers', 'bases', 'default_access_level', 'g', '_pro', '__weakref__',
    
    name: str
    bases: ProEntries
    default_access_level: AccessLevel 
    g: ReadonlyDict['DepGraph', 'DepGraph']
    providers: ReadonlyDict[Injectable, Provider]
    _pro: FrozenDict[Self, int]
    is_atomic: t.Final = True
    
    def __init__(self, name: str=None, *bases: Self, module: str, access_level: AccessLevel=PUBLIC) -> None:
        """Create a container.
        
        Params:
            name (str, optional): Name of the container
            *bases (Container, optional): Base container.
            access_level (AccessLevel, optional): The default `access_level`
                to assign providers
        """
        if name and not name.isidentifier():
            raise ValueError(f'name must be a valid identifier not {name!r}')
        
        self.__setattr(
            _pro=None, 
            bases=ProEntries(),
            name=name or f'__anonymous__', 
            providers=ReadonlyDict(),
            module=module, 
            g=ReadonlyDict(),
            default_access_level=AccessLevel(access_level)
        )
        
        bases and self.extend(*bases)
        signals.on_container_create.send(self.__class__, container=self)


    @property
    def atomic(self):
        """`AtomicProEntries`(s) 
        """
        return AtomicProEntries(((self, None),))

    def extend(self, *bases: Self) -> Self:
        """Adds containers to extended by this container.
        Args:
            *bases (Container): The base containers to be extended
            
        Returns:
            Self: this container
        """
        self.__setattr(bases=self.bases | ProEntries.make(bases))
        return self

    def access_level(self, accessor: Self):
        """Get the `AccessLevel` 

        Params:
            accessor (Container): 

        Returns:
            access_level (AccessLevel):
        """
        if accessor is self:
            return PRIVATE
        elif self.extends(accessor):
            return GUARDED
        elif accessor.extends(self):
            return PROTECTED
        else:
            return PUBLIC

    @classmethod
    def _collect(cls, *a, **kw) -> 'Group': 
        return Group(*a, **kw)

    def _on_register(self, abstract: Injectable, provider: Provider):
        pass

    def __contains__(self, x):
        return x in self.providers or any(x in b for b in self.bases)

    def __setitem__(self, key: Injectable, provider: Provider) -> Self:
        """Register a dependency provider 
        
            container[_T] = providers.Value('abc')

        Params:
            abstract (Injectable): The dependency to be provided
            provider (Provider): The provider to provide the dependency
        """
        if not is_injectable(key):
            raise TypeError(f'expected `Injectable` not. `{key.__class__.__qualname__}`')

        if prov := provider._setup(self, key):
            self._on_register(key, prov)
            _dict_setitem(self.providers, key, prov)
            # self.__setitem(key, prov)
            signals.on_provider_registered.send(self, abstract=key, provider=provider)

    def __getitem__(self, k):
        try:
            return self.providers[k]    
        except KeyError:
            if isinstance(k, Provider) and (k.container or self) is self:
                return k
        
    def _resolve(self, key: 'DepKey', bindings: 'DepGraph'):
        if prov := self[key.abstract]:
            access = prov.access_level or self.default_access_level

            if access in self.access_level(key.container):
                if prov._can_resolve(key, bindings):
                    return prov,
        return ()

    def __bool__(self):
        return True
        


class Group(BaseContainer):
    """A `Container` group.
    """
    __slots__ = 'g', 'bases', 'name', 'module', '_pro', '__weakref__',
    is_atomic = False
    
    bases: ProEntries
    _pro: FrozenDict[Self, int]
    
    @t.overload
    def __new__(cls: type[Self], it: t.Union[abc.Iterable[Container], Self]=(), *, name: str=None, module: str=None) -> Self: ...
    def __new__(cls: type[Self], it: t.Union[abc.Iterable[Container], Self]=(), *, module: str, **kwds) -> Self:
        self = _object_new(cls)
        typ: type[it] = it.__class__
        if issubclass(typ, cls):
            if not kwds and module == it.module:
                return it
            else:
                kwds['bases'], kwds['_pro'] = it.bases, it._pro
        elif typ is ProEntries:
            kwds['bases'], kwds['_pro'] = it, None
        else:
            kwds['bases'] = it = ProEntries.make(it)
            kwds['_pro'] = None

        kwds.setdefault('name', None)
        self.__setattr(module=module, g=ReadonlyDict(), **kwds)
        return self

    @property
    def atomic(self):
        return self.bases.atomic()

    @property
    def qualname(self) -> None:
        return f'{self.module}:{self.name or "|".join(ordered_set(c.qualname for c in self.bases))}'

    @property
    def providers(self):
        return ChainMap(*(a.providers for a in self.bases))

    @classmethod
    def _collect(cls, *a, **kw): 
        return cls(*a, **kw)

    def __sub__(self, o):
        if isinstance(o, BaseContainer):
            check = o.atomic
            return self._collect(x for x in self.atomic if not x in check)
        return NotImplemented
    __isub__ = __sub__
    
    # def __rsub__(self, o):
    #     if isinstance(o, Group):
    #         check = self.atomic
    #         return self._collect(x for x in o.atomic if not x in check)
    #     return NotImplemented
    
    def __bool__(self):
        return not not self.bases

    def __contains__(self, x):
        return any(x in b for b in self.bases)

    

ContainerMeta.register = None



from .graph import DepGraph, _null_graph




from ._common import Missing


from .markers import (
    Injectable, is_injectable, T_Injected, T_Injectable,
    Dep, DependencyMarker, Lookup, PureDep
)


from .exceptions import InjectorLookupError, UziException


from . import injectors, providers
from .containers import Container
from .injectors import Injector
from .providers import Provider
from .scopes import Scope, ThreadSafeScope, ThreadLocalScope, ContextLocalScope


from . import _receivers





from abc import ABC, abstractmethod
import logging
from threading import Lock
from typing_extensions import Self
import attr
import typing as t 

from collections.abc import Callable

from ._common import Missing, private_setattr
from ._functools import BoundParams, _PositionalArgs, _PositionalDeps, _KeywordDeps, FutureFactoryWrapper, FutureResourceWrapper, FutureCallableWrapper

from .markers import T_Injectable, T_Injected
from .exceptions import InjectorLookupError

if t.TYPE_CHECKING: # pragma: no cover
    from .providers import Provider
    from .graph import DepGraph
    from .injectors import Injector
    from .containers import Container


_object_new = object.__new__


_T_Concrete = t.TypeVar("_T_Concrete")
"""Provided `concrete` `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
@private_setattr
class Binding(ABC, t.Generic[_T_Concrete]):
    """Marks binding an injectable as a `dependency` to be injected."""
    
    @abstractmethod
    def bind(self, injector: 'Injector') -> t.Union[Callable[..., T_Injected], None]: 
        raise NotImplementedError(f'{self.__class__.__name__}.bind()')  # pragma: no cover

    abstract: T_Injectable = attr.ib()
    graph: "DepGraph" = attr.ib()
    provider: "Provider" = attr.ib(default=None, repr=lambda p: str(p and id(p)))

    concrete: _T_Concrete = attr.ib(kw_only=True, default=Missing, repr=True)

    is_async: bool = False
    dependencies = frozenset[Self]()

    _v_ident: tuple = attr.ib(init=False, repr=False)
    @_v_ident.default
    def _init_v_ident(self):
        return self.abstract, self.graph, self.container

    _ash: tuple = attr.ib(init=False, repr=False)
    @_ash.default
    def _init_ash(self):
        return hash(self._v_ident)

    @property
    def container(self):
        if pro := self.provider or self.graph:
            return pro.container

    def __eq__(self, o: Self) -> bool:
        if o.__class__ is self.__class__:
            return o._v_ident == self._v_ident
        elif isinstance(o, Binding):
            return False
        return NotImplemented

    def __ne__(self, o: Self) -> bool:
        if o.__class__ is self.__class__:
            return o._v_ident != self._v_ident
        elif isinstance(o, Binding):
            return True
        return NotImplemented

    def __hash__(self) -> int:
        return self._ash




_T_Binding = t.TypeVar('_T_Binding', bound=Binding, covariant=True)
"""Dependency `TypeVar`"""




@private_setattr
class LookupErrorBinding:

    __slots__ = 'abstract', 'graph',

    graph: 'DepGraph'
    container = None
    is_async: bool = False
    dependencies = frozenset[Self]()

    def __new__(cls: type[Self], abstract=None, graph: 'DepGraph'=None, provider=None, concrete=None) -> Self:
        self = _object_new(cls)
        self.__setattr(abstract=abstract, graph=graph)
        return self

    def __bool__(self): 
        return False

    def __reduce__(self): 
        return self.__class__, (self.abstract, self.graph)

    def __eq__(self, o):
        return self.abstract == o

    def __ne__(self, o):
        return not self == o

    def __hash__(self) -> int:
        return hash(self.abstract)

    def bind(self, injector: 'Injector'):
        raise InjectorLookupError(self.abstract, self.graph or injector.graph)




@attr.s(slots=True, frozen=True, cmp=False)
class SimpleBinding(Binding[_T_Concrete]):

    def bind(self, injector: 'Injector'):
        return self.concrete(injector)






@attr.s(slots=True, frozen=True, cmp=False)
class Value(Binding[T_Injected]):
    """Value binding
    """

    concrete: T_Injected = attr.ib(kw_only=True, default=None)
    is_async: t.Final = False

    def bind(self, injector: 'Injector'):
        return self

    def __call__(self) -> T_Injected:
        return self.concrete




_T_ValueBinding = t.TypeVar('_T_ValueBinding', bound=Value, covariant=True)
"""Value binding `TypeVar`"""




@attr.s(slots=True, frozen=True, cmp=False)
class Factory(Binding[T_Injected]):
    """Factory binding"""

    concrete: T_Injected = attr.ib(kw_only=True)
    params: 'BoundParams' = attr.ib(kw_only=True, default=BoundParams.make(()))
    thread_safe: bool = attr.ib(kw_only=True, default=False)

    @property
    def dependencies(self):
        return self.params.dependencies

    def bind(self: Self, injector: "Injector"):
        if self.params:
            args = self.resolve_args(injector)
            kwargs = self.resolve_kwargs(injector)
            vals = self.params.vals
            func = self.concrete
            def factory():
                nonlocal func, args, kwargs, vals
                return func(*args, **kwargs, **vals)
            return factory
        else:
            return self.concrete

    def resolve_args(self, injector: "Injector"):
        params = self.params
        if params.args:
            if params._pos_vals > 0 < params._pos_deps:
                return _PositionalArgs(
                    (p.value, None) if p.has_value else (None, injector[p.dependency])
                    for p in params.args
                )
            elif params._pos_deps > 0:
                return _PositionalDeps(
                    injector[p.dependency] for p in params.args
                )
            else:
                return tuple(p.value for p in params.args)
        return ()

    def resolve_kwargs(self, injector: "Injector"):
        return _KeywordDeps((p.key, injector[p.dependency]) for p in self.params.kwds)




_T_FactoryBinding = t.TypeVar('_T_FactoryBinding', bound=Factory, covariant=True)
"""Factory binding `TypeVar`"""



@attr.s(slots=True, frozen=True, cmp=False)
class AsyncFactory(Factory[T_Injected]):
    """AsyncFactory binding"""

    is_async: bool = True
    async_call: bool = True




@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsFactory(Factory[T_Injected]):
    """AwaitParamsFactory binding"""
    
    is_async: bool = True
    async_call: bool = False
     
    def bind(self: Self, injector: "Injector"):
        args, aw_args = self.resolve_args(injector)
        kwargs, aw_kwargs = self.resolve_kwargs(injector)
        return FutureFactoryWrapper(
            self.concrete, self.params.vals, 
            args=args, kwargs=kwargs, 
            aw_args=aw_args, aw_kwargs=aw_kwargs,
            aw_call=self.async_call, 
        )

    def resolve_kwargs(self, injector: "Injector"):
        if self.params.aw_kwds:
            deps = super().resolve_kwargs(injector)
            return deps, tuple((n, deps.pop(n)) for n in self.params.aw_kwds)
        else:
            return super().resolve_kwargs(injector), ()

    def resolve_args(self, injector: "Injector"):
        return super().resolve_args(injector), self.params.aw_args



@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncFactory(AwaitParamsFactory[T_Injected]):
    """AwaitParamsAsyncFactory binding"""
    
    async_call: bool = True





@attr.s(slots=True, frozen=True, cmp=False)
class Singleton(Factory[T_Injected]):
    """Singleton binding"""

    # aw_enter: bool = attr.ib(kw_only=True, default=False)

    def factory(self, injector: 'Injector'):
        if self.params:
            args = self.resolve_args(injector)
            kwargs = self.resolve_kwargs(injector)
            vals = self.params.vals
            func = self.concrete
            return lambda: func(*args, **kwargs, **vals)
        else:
            return self.concrete

    def bind(self, injector: 'Injector'):
        func = self.factory(injector)
        value = Missing
        lock = Lock() if self.thread_safe else None
        
        def factory():
            nonlocal func, value
            if not value is Missing:
                return value

            lock and lock.acquire(blocking=True)
            try:
                if value is Missing:
                    value = func()
            finally:
                lock and lock.release()
            return value

        return factory


_T_SingletonBinding = t.TypeVar('_T_SingletonBinding', bound=Singleton, covariant=True)
"""Singleton binding `TypeVar`"""



@attr.s(slots=True, frozen=True, cmp=False)
class AsyncSingleton(Singleton[T_Injected]):
    """AsyncSingleton binding"""

    is_async: bool = True
    async_call: bool = True

    def factory(self, injector: 'Injector'):
        if params := self.params:
            args, kwargs = self.resolve_args(injector), self.resolve_kwargs(injector)
            return FutureFactoryWrapper(self.concrete, params.vals, args=args, kwargs=kwargs, aw_call=self.async_call)
        else:
            return FutureFactoryWrapper(self.concrete, self.params.vals, aw_call=self.async_call)




@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsSingleton(Singleton[T_Injected], AwaitParamsFactory[T_Injected]):
    """AwaitParamsSingleton binding"""

    is_async: bool = True
    async_call: bool = False

    def factory(self, injector: 'Injector'):
        (args, aw_args), (kwargs, aw_kwargs) = self.resolve_args(injector), self.resolve_kwargs(injector)
        return FutureFactoryWrapper(self.concrete, self.params.vals, args=args, kwargs=kwargs, aw_args=aw_args, aw_kwargs=aw_kwargs, aw_call=self.async_call)


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncSingleton(AwaitParamsSingleton[T_Injected]):
    """AwaitParamsAsyncSingleton binding"""
    
    async_call: bool = True



@attr.s(slots=True, frozen=True, cmp=False)
class Resource(Singleton[T_Injected]):
    """Binds resources.
    """


_T_ResourceBinding = t.TypeVar('_T_ResourceBinding', bound=Resource, covariant=True)
"""Resource binding `TypeVar`"""





@attr.s(slots=True, frozen=True, cmp=False)
class Partial(Factory[T_Injected]):
    """Partial binding"""

    def factory(self: Self, injector: "Injector"):
        args = self.resolve_args(injector)
        kwargs = self.resolve_kwargs(injector)
        vals = self.params.vals
        func = self.concrete

        def make(*a, **kw):
            nonlocal func, args, kwargs, vals
            return func(*args, *a, **(vals | kw), **kwargs.skip(kw))

        return make

    def bind(self: Self, injector: "Injector"):
        return self.factory(injector)



_T_PartialBinding  = t.TypeVar('_T_PartialBinding', bound=Partial, covariant=True)
"""Partial binding `TypeVar`"""



@attr.s(slots=True, frozen=True, cmp=False)
class AsyncPartial(Partial[T_Injected]):
    """AsyncPartial binding"""

    async_call: bool = True
    is_async: bool = True



@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsPartial(Partial[T_Injected], AwaitParamsFactory[T_Injected]):
    """AwaitParamsPartial binding"""
    
    async_call: bool = False
    is_async: bool = True
    
    def factory(self: Self, injector: 'Injector'):
        (args, aw_args), (kwargs, aw_kwargs) = self.resolve_args(injector), self.resolve_kwargs(injector)
        return FutureCallableWrapper(self.concrete, self.params.vals, args=args, kwargs=kwargs, aw_args=aw_args, aw_kwargs=aw_kwargs, aw_call=self.async_call)



@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncPartial(AwaitParamsPartial[T_Injected]):
    """AwaitParamsAsyncPartial binding"""
    
    async_call: bool = True






@attr.s(slots=True, frozen=True, cmp=False)
class Callable(Partial[T_Injected]):
    """Callable binding"""

    def bind(self: Self, injector: "Injector"):
        func = self.factory(injector)
        return lambda: func


_T_CallableBinding = t.TypeVar('_T_CallableBinding', bound=Callable, covariant=True)
"""Callable binding `TypeVar`"""



@attr.s(slots=True, frozen=True, cmp=False)
class AsyncCallable(Callable[T_Injected], AsyncPartial[T_Injected]):
    """AsyncCallable binding"""
    is_async: bool = True



@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsCallable(Callable[T_Injected], AwaitParamsPartial[T_Injected]):
    """AwaitParamsCallable binding"""
    
    is_async: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncCallable(Callable[T_Injected], AwaitParamsAsyncPartial[T_Injected]):
    """AwaitParamsAsyncCallable binding"""
    
    async_call: bool = True



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False


import typing as t
from asyncio import AbstractEventLoop, Future, ensure_future, get_running_loop
from collections.abc import Callable, ItemsView, Iterator, Mapping, ValuesView
from contextlib import AbstractAsyncContextManager
from inspect import Parameter, Signature
from logging import getLogger

import attr
from typing_extensions import Self

from uzi._common import FrozenDict

from ._common import Missing
from .markers import Injectable, is_injectable_annotation
from .markers import DependencyMarker

if t.TYPE_CHECKING:  # pragma: no cover
    from ._bindings import Binding
    from .containers import Container
    from .injectors import Injector
    from .graph import DepGraph


logger = getLogger(__name__)

_POSITIONAL_ONLY = Parameter.POSITIONAL_ONLY
_VAR_POSITIONAL = Parameter.VAR_POSITIONAL
_POSITIONAL_KINDS = frozenset([_POSITIONAL_ONLY, _VAR_POSITIONAL])
_POSITIONAL_OR_KEYWORD = Parameter.POSITIONAL_OR_KEYWORD
_KEYWORD_ONLY = Parameter.KEYWORD_ONLY
_VAR_KEYWORD = Parameter.VAR_KEYWORD
_KEYWORD_KINDS = frozenset(
    [Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD]
)

_EMPTY = Parameter.empty

_T = t.TypeVar("_T")


_frozendict = FrozenDict()


_object_new = object.__new__
_object_setattr = object.__setattr__


class BoundParam:
    """A bound param    
    """


    __slots__ = (
        "param",
        "key",
        "value",
        "injectable",
        "dependency",
        "has_default",
    )

    param: Parameter
    name: str
    annotation: t.Any
    value: _T
    default: t.Any
    injectable: Injectable
    dependency: "Binding"

    def __new__(
        cls, param: Parameter, value: t.Any = _EMPTY, key: t.Union[str, int] = None
    ):
        self = object.__new__(cls)
        self.param = param
        self.key = key or param.name
        self.dependency = self.injectable = None
        self.has_default = False

        if isinstance(value, DependencyMarker):
            self.injectable = value
        elif not value is _EMPTY:
            self.value = value

        if isinstance(param.default, DependencyMarker):
            if None is self.injectable:
                self.injectable = param.default
        else:
            self.has_default = not param.default is _EMPTY

        if None is self.injectable:
            annotation = param.annotation
            if not annotation is _EMPTY and is_injectable_annotation(annotation):
                self.injectable = annotation

        return self

    @property
    def name(self):
        return self.param.name

    @property
    def is_async(self):
        if dep := self.dependency:
            return dep.is_async

    @property
    def is_injectable(self):
        return not self.injectable is None

    @property
    def has_value(self):
        return hasattr(self, "value")

    @property
    def default(self):
        return self.param.default

    @property
    def annotation(self):
        return self.param.annotation

    @property
    def kind(self):
        return self.param.kind


@attr.s(slots=True, frozen=True)
class BoundParams:
    """A collection of bound params
    """

    params: tuple[BoundParam] = attr.ib(converter=tuple)

    args: tuple[BoundParam] = attr.ib(converter=tuple, kw_only=True)
    aw_args: tuple[int] = attr.ib(converter=tuple)
    kwds: tuple[BoundParam] = attr.ib(converter=tuple)
    aw_kwds: tuple[str] = attr.ib(converter=tuple)
    is_async: bool = attr.ib()
    vals: FrozenDict[str, t.Any] = attr.ib(converter=FrozenDict)
    _pos_vals: int = attr.ib(converter=int)
    _pos_deps: int = attr.ib(converter=int)

    @property
    def dependencies(self) -> set["Binding"]:
        return dict.fromkeys(p.dependency for p in self.params if p.dependency).keys()

    @classmethod
    def make(cls, params: tuple[BoundParam]) -> None:
        args = []
        kwds = []
        vals = {}
        aw_args = []
        aw_kwds = []
        pos_vals = pos_deps = i = 0
        skip_pos = False

        params = tuple(params)
        for p in params:
            if p.kind in _POSITIONAL_KINDS:
                if skip_pos is False:
                    if p.has_value or p.dependency:
                        args.append(p)
                        p.is_async and aw_args.append(i)
                        pos_vals += p.has_value
                        pos_deps += not not p.dependency and not p.has_value
                    else:
                        skip_pos = True
                    i += 1
                continue
            elif p.has_value:
                vals[p.key] = p.value
            elif p.dependency:
                kwds.append(p)
                p.is_async and aw_kwds.append(p.key)
        return cls(
            params=params,
            args=args,
            kwds=kwds,
            vals=vals,
            pos_vals=pos_vals,
            pos_deps=pos_deps,
            aw_args=aw_args,
            aw_kwds=aw_kwds,
            is_async=not not (aw_args or aw_kwds),
        )

    @classmethod
    def bind(
        cls,
        sig: Signature,
        scope: "DepGraph" = None,
        container: "Container" = None,
        args: tuple = (),
        kwargs: dict = FrozenDict(),
    ) -> Self:
        return cls.make(cls._iter_bind(sig, scope, container, args, kwargs))

    @classmethod
    def _iter_bind(
        cls,
        sig: Signature,
        scope: "DepGraph" = None,
        container: "Container" = None,
        args=(),
        kwargs=FrozenDict(),
    ):
        bound = sig.bind_partial(*args, **kwargs).arguments
        container = container or scope.container
        for n, p in sig.parameters.items():
            if p.kind is Parameter.VAR_POSITIONAL:
                p = p.replace(annotation=Parameter.empty)
                for v in bound.get(n) or (Parameter.empty,):
                    bp = BoundParam(p, v)
                    if scope and bp.is_injectable:
                        bp.dependency = scope[bp.injectable]
                    yield bp
            elif p.kind is Parameter.VAR_KEYWORD:
                p = p.replace(annotation=Parameter.empty)
                for k, v in (bound.get(n) or {n: Parameter.empty}).items():
                    bp = BoundParam(p, v, key=k)
                    if scope and bp.is_injectable:
                        bp.dependency = scope[bp.injectable]
                    yield bp
            else:
                bp = BoundParam(p, bound.get(n, Parameter.empty))
                if scope and bp.is_injectable:
                    bp.dependency = scope[bp.injectable]
                yield bp

    def __bool__(self):
        return not not self.params


class _PositionalArgs(tuple[tuple[t.Any, Callable[[], _T]]], t.Generic[_T]):

    __slots__ = ()

    __raw_new__ = classmethod(tuple.__new__)

    # def __reduce__(self): return tuple, (tuple(self),)

    def copy(self): return self[:]

    __copy__ = copy

    @t.overload
    def __getitem__(self, index: int) -> tuple[_T, bool]: ...

    @t.overload
    def __getitem__(self, slice: slice) -> Self: ...

    def __getitem__(self, index: t.Union[int, slice]) -> t.Union[tuple[_T, bool], Self]:
        v, fn = self.get_raw(index)
        if not fn:
            return v
        return fn()

    def __iter__(self) -> "Iterator[tuple[_T, bool]]":
        for v, fn in self.iter_raw():
            if not fn:
                yield v
            else:
                yield fn()

    if t.TYPE_CHECKING:  # pragma: no cover

        def get_raw(index: int) -> tuple[t.Any, Callable[[], _T]]:
            ...

        def iter_raw() -> Iterator[tuple[t.Any, Callable[[], _T]]]:
            ...

    else:
        get_raw = tuple.__getitem__
        iter_raw = tuple.__iter__


class _PositionalDeps(_PositionalArgs):
    __slots__ = ()

    def __getitem__(self, index: t.Union[int, slice]) -> t.Union[tuple[_T, bool], Self]:
        return self.get_raw(index)()

    def __iter__(self) -> "Iterator[tuple[_T, bool]]":
        for yv in self.iter_raw():
            yield yv()


class _KeywordDeps(dict[str, Callable[[], _T]], t.Generic[_T]):

    __slots__ = ()

    def __getitem__(self, name: str):
        return self.get_raw(name)()

    def __iter__(self) -> "Iterator[_T]":
        return self.iter_raw()

    def __reduce__(self):
        return dict, (tuple(self.items()),)

    def copy(self):
        return self.__class__(self)

    __copy__ = copy

    def items(self) -> "ItemsView[tuple[str, _T]]":
        return ItemsView(self)

    def values(self) -> "ValuesView[_T]":
        return ValuesView(self)

    def skip(self, skip: Mapping[str, t.Any]):
        if skip:
            return {k: self.get_raw(k)() for k in self.iter_raw() if not k in skip}
        else:
            return self

    iter_raw = dict.__iter__
    get_raw = dict.__getitem__
    raw_items = dict.items
    raw_values = dict.values


class FutureFactoryWrapper:

    __slots__ = (
        "_func",
        "_args",
        "_kwargs",
        "_vals",
        "_aw_args",
        "_aw_kwargs",
        "_aw_call",
    )

    _func: Callable
    _args: "_PositionalArgs"
    _kwargs: "_KeywordDeps"
    _vals: Mapping

    def __new__(
        cls,
        func,
        vals: Mapping = FrozenDict(),
        args: "_PositionalArgs" = FrozenDict(),
        kwargs: "_KeywordDeps" = FrozenDict(),
        *,
        aw_args: tuple[int] = FrozenDict(),
        aw_kwargs: tuple[str] = FrozenDict(),
        aw_call: bool = True,
    ) -> Self:
        self = _object_new(cls)
        self._func = func
        self._vals = vals
        self._args = args
        self._kwargs = kwargs
        self._aw_args = aw_args
        self._aw_kwargs = aw_kwargs
        self._aw_call = aw_call
        return self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}:{self._func.__module__}.{self._func.__qualname__}()"

    def __call__(self):
        loop = get_running_loop()
        if aw_args := self._aw_args:
            _args = self._args
            aw_args = {i: ensure_future(_args[i], loop=loop) for i in aw_args}
        if aw_kwargs := self._aw_kwargs:
            aw_kwargs = {n: ensure_future(d(), loop=loop) for n, d in aw_kwargs}

        return FactoryFuture(self, aw_args, aw_kwargs, loop=loop)


class FutureCallableWrapper(FutureFactoryWrapper):

    __slots__ = ()

    def __call__(self, *a, **kw):
        loop = get_running_loop()
        if aw_args := self._aw_args:
            _args = self._args
            aw_args = {i: ensure_future(_args[i], loop=loop) for i in aw_args}
        if aw_kwargs := self._aw_kwargs:
            aw_kwargs = {
                n: ensure_future(d(), loop=loop) for n, d in aw_kwargs if not n in kw
            }

        return CallableFuture(self, aw_args, aw_kwargs, args=a, kwargs=kw, loop=loop)


class FutureResourceWrapper(FutureFactoryWrapper):  # pragma: no cover

    __slots__ = (
        "_aw_enter",
        "_ctx",
    )

    _aw_enter: bool
    _ctx: "Injector"

    if not t.TYPE_CHECKING:  # pragma: no cover

        def __new__(cls, ctx, *args, aw_enter: bool = None, **kwargs):
            self = super().__new__(cls, *args, **kwargs)
            self._ctx = ctx
            self._aw_enter = aw_enter
            return self

    def __call__(self, *a, **kw):
        loop = get_running_loop()
        if aw_args := self._aw_args:
            _args = self._args
            aw_args = {i: ensure_future(_args[i], loop=loop) for i in aw_args}
        if aw_kwargs := self._aw_kwargs:
            aw_kwargs = {n: ensure_future(d(), loop=loop) for n, d in aw_kwargs}

        return ResourceFuture(self, aw_args, aw_kwargs, loop=loop)


class FactoryFuture(Future):

    _asyncio_future_blocking = False
    _loop: AbstractEventLoop
    _factory: FutureFactoryWrapper
    _aws: tuple[dict[int, Future[_T]], dict[str, Future[_T]]]
    # _aws: tuple[Future[_T], dict[str, Future[_T]]]

    def __init__(
        self, factory, aw_args=FrozenDict(), aw_kwargs=FrozenDict(), *, loop=None
    ) -> Self:
        Future.__init__(self, loop=loop)
        self._factory = factory
        self._aws = aw_args, aw_kwargs

    def __await__(self):
        if not self.done():
            self._asyncio_future_blocking = True
            factory = self._factory
            aw_args, aw_kwargs = self._aws

            if aw_args:
                for k in aw_args:
                    aw_args[k] = yield from aw_args[k]
                _args = factory._args
                args = (
                    aw_args[i] if i in aw_args else _args[i] for i in range(len(_args))
                )
            else:
                args = factory._args

            if aw_kwargs:
                for k in aw_kwargs:
                    aw_kwargs[k] = yield from aw_kwargs[k]
            else:
                aw_kwargs = _frozendict

            res = factory._func(*args, **aw_kwargs, **factory._kwargs, **factory._vals)
            if factory._aw_call:
                res = yield from ensure_future(res, loop=self._loop)
            self.set_result(res)  # res
            return res
        return self.result()

    __iter__ = __await__


class CallableFuture(Future):

    _asyncio_future_blocking = False
    _loop: AbstractEventLoop
    _factory: FutureCallableWrapper
    _aws: tuple[dict[int, Future[_T]], dict[str, Future[_T]]]
    # _aws: tuple[Future[_T], dict[str, Future[_T]]]
    _extra_args: tuple[t.Any]
    _extra_kwargs: Mapping[str, t.Any]

    def __init__(
        self,
        factory,
        aw_args=FrozenDict(),
        aw_kwargs=FrozenDict(),
        *,
        args: tuple = (),
        kwargs: dict[str, t.Any] = FrozenDict(),
        loop=None,
    ) -> Self:
        Future.__init__(self, loop=loop)
        self._factory = factory
        self._aws = aw_args, aw_kwargs
        self._extra_args = args
        self._extra_kwargs = kwargs

    def __await__(self):
        if not self.done():
            self._asyncio_future_blocking = True
            factory = self._factory
            aw_args, aw_kwargs = self._aws

            if aw_args:
                for k in aw_args:
                    aw_args[k] = yield from aw_args[k]
                _args = factory._args
                args = (
                    aw_args[i] if i in aw_args else _args[i] for i in range(len(_args))
                )
            else:
                args = factory._args

            if aw_kwargs:
                for k, aw in aw_kwargs.items():
                    aw_kwargs[k] = yield from aw_kwargs[k]

            if kwargs := self._extra_kwargs:
                vals = factory._vals | kwargs
                kwargs = factory._kwargs.skip(kwargs)
            else:
                vals = factory._vals
                kwargs = factory._kwargs

            res = factory._func(*args, *self._extra_args, **aw_kwargs, **kwargs, **vals)
            if factory._aw_call:
                res = yield from ensure_future(res, loop=self._loop)
            self.set_result(res)  # res
            return res
        return self.result()

    __iter__ = __await__


class ResourceFuture(Future):  # pragma: no cover

    _asyncio_future_blocking = False
    _loop: AbstractEventLoop
    _factory: FutureResourceWrapper
    _aws: tuple[dict[int, Future[_T]], dict[str, Future[_T]]]
    # _aws: tuple[Future[_T], dict[str, Future[_T]]]
    _extra_args: tuple[t.Any]
    _extra_kwargs: Mapping[str, t.Any]

    def __init__(
        self,
        factory,
        aw_args=FrozenDict(),
        aw_kwargs=FrozenDict(),
        *,
        args: tuple = (),
        kwargs: dict[str, t.Any] = FrozenDict(),
        loop=None,
    ) -> Self:
        Future.__init__(self, loop=loop)
        self._factory = factory
        self._aws = aw_args, aw_kwargs
        self._extra_args = args
        self._extra_kwargs = kwargs

    def __await__(self):
        if not self.done():
            self._asyncio_future_blocking = True
            factory = self._factory
            aw_args, aw_kwargs = self._aws

            if aw_args:
                for k in aw_args:
                    aw_args[k] = yield from aw_args[k]
                _args = factory._args
                args = (
                    aw_args[i] if i in aw_args else _args[i] for i in range(len(_args))
                )
            else:
                args = factory._args

            if aw_kwargs:
                for k, aw in aw_kwargs.items():
                    aw_kwargs[k] = yield from aw_kwargs[k]

            if kwargs := self._extra_kwargs:
                vals = factory._vals | kwargs
                kwargs = factory._kwargs.skip(kwargs)
            else:
                vals = factory._vals
                kwargs = factory._kwargs

            res = factory._func(*args, *self._extra_args, **aw_kwargs, **kwargs, **vals)
            if factory._aw_call:
                res = yield from ensure_future(res, loop=self._loop)

            aw_enter = factory._aw_enter
            cm = factory._ctx.exitstack
            if aw_enter is False:
                res = cm.enter(res)
            elif aw_enter is True:
                res = yield from self._loop.create_task(cm.enter_async(res))
            elif isinstance(res, AbstractAsyncContextManager):
                factory._aw_enter = True
                res = yield from self._loop.create_task(cm.enter_async(res))
            else:
                factory._aw_enter = False
                res = cm.enter(res)

            self.set_result(res)  # res
            return res
        return self.result()

    __iter__ = __await__



from functools import reduce, wraps
from inspect import Parameter, signature
from logging import getLogger
import operator
from types import FunctionType, GenericAlias, MethodType
import typing as t
from abc import ABC, ABCMeta, abstractmethod
from collections import abc
from enum import Enum

from typing_extensions import Self


from ._common import Missing, private_setattr
from ._common.lookups import Lookup as BaseLookup


if t.TYPE_CHECKING: # pragma: no cover
    from .containers import Container
    from .graph import DepGraph, DepSrc




_object_new = object.__new__



T_Injected = t.TypeVar("T_Injected", covariant=True)
"""The injected type.
"""

T_Default = t.TypeVar("T_Default")
"""Default value type.
"""

T_Injectable = t.TypeVar("T_Injectable", bound="Injectable", covariant=True)
"""An `Injectable` type.
"""

logger = getLogger(__name__)

_NoneType = type(None)


_BLACKLIST = frozenset(
    {
        None,
        _NoneType,
        t.Any,
        type(t.Literal[1]),
        str,
        bytes,
        bytearray,
        tuple,
        int,
        float,
        frozenset,
        set,
        dict,
        list,
        Parameter.empty,
        Missing,
    }
)


def is_injectable(obj):
    """Returns `True` if the given type annotation is injectable.
    
    Params: 
        typ (type): The type annotation to check.
    Returns:
        (bool): `True` if `typ` can be injected or `False` if otherwise.
    """
    return isinstance(obj, Injectable) and not (
        obj in _BLACKLIST or isinstance(obj, NonInjectable)
    )


def is_injectable_annotation(typ):
    """Returns `True` if the given type annotation is injectable.
    
    Params: 
        typ (type): The type annotation to check.
    Returns:
        (bool): `True` if `typ` can be injected or `False` if otherwise.
    """
    return is_injectable(typ)



class Injectable(metaclass=ABCMeta):
    """Abstract base class for injectable types.

    An injectable is an object that can be used to represent a dependency.
    
    Builtin injectable types:- `type`, `TypeVar`, `FunctionType`, `MethodType`, 
    `GenericAlias`
    """
    __slots__ = ()


Injectable.register(type)
Injectable.register(t.TypeVar)
Injectable.register(FunctionType)
Injectable.register(MethodType)
Injectable.register(GenericAlias)
Injectable.register(type(t.Generic[T_Injected]))
Injectable.register(type(t.Union))


class NonInjectable(metaclass=ABCMeta):
    """Abstract base class for non-injectable types."""
    __slots__ = ()


NonInjectable.register(_NoneType)
NonInjectable.register(type(t.Literal[1]))



__static_makers = {
    t.Union,
    t.Annotated,
}


@t.overload
def is_dependency_marker(obj: 'DependencyMarker') -> True: ...
@t.overload
def is_dependency_marker(obj: object) -> False: ...
def is_dependency_marker(obj:t.Any, ind=0) -> bool:
    """Check if object is a `DependencyMarker`
    Args:
        obj (Any): the object to check

    Returns:
        bool: 
    """
    return isinstance(obj, (DependencyMarker, DependencyMarkerType)) \
            or obj in __static_makers \
            or (not not (orig := t.get_origin(obj)) and is_dependency_marker(orig))



class DependencyMarkerType(ABCMeta):
    ...


class DependencyMarker(Injectable, t.Generic[T_Injectable], metaclass=DependencyMarkerType):
    """Abstract base class for dependency markers. 

    Dependency markers are used reperesent and/or annotate dependencies. 
    """
    __slots__ = ()

    @property
    @abstractmethod
    def __origin__(self): ...





_3_nones = None, None, None

_T_Start = t.TypeVar('_T_Start', int, 'ProPredicate', None)
_T_Stop = t.TypeVar('_T_Stop', int, 'ProPredicate', None)
_T_Step = t.TypeVar('_T_Step', int, None)

_T_PredVar = t.TypeVar('_T_PredVar')
_T_PredVars = t.TypeVar('_T_PredVars', bound=tuple)
_T_Pred = t.TypeVar('_T_Pred', bound='ProPredicate', covariant=True)



@private_setattr
class _PredicateBase:

    __slots__ = 'vars',

    vars: tuple[_T_PredVar]

    def __new__(cls, *vars: _T_PredVar):
        self = _object_new(cls)
        self.__setattr(vars=vars)
        return self

    @abstractmethod
    def pro_entries(self, it: abc.Iterable['Container'], graph: 'DepGraph', src: 'DepSrc') -> abc.Iterable['Container']:  # pragma: no cover
        raise NotImplementedError(f'{self.__class__.__qualname__}.pro_entries()')

    def __copy__(self):
        return self.__class__(*self.vars)

    def __reduce__(self):
        return self.__class__, tuple(self.vars)


    
@private_setattr
class _PredicateOpsMixin:
    
    __slots__ = ()

    def __or__(self, x):    
        if isinstance(x, ProPredicate):
            if x == self:
                return self
            return ProOrPredicate(self, x)
        return NotImplemented
    __ior__ = __or__
    
    def __and__(self, x):
        if isinstance(x, ProPredicate):
            if x == self:
                return self
            return ProAndPredicate(self, x)
        return NotImplemented
    __iand__ = __and__
    
    def __ror__(self, x):    
        if isinstance(x, ProPredicate):
            return ProOrPredicate(x, self)
        return NotImplemented

    def __rand__(self, x):
        if isinstance(x, ProPredicate):
            return ProAndPredicate(x, self)
        return NotImplemented
   
    def __invert__(self):
        return ProInvertPredicate(self)



class _PredicateCompareMixin:
    
    __slots__ = ()

    def __eq__(self, o) -> bool:
        if isinstance(o, self.__class__):
            return self.vars == o.vars
        return NotImplemented

    def __ne__(self, o) -> bool:
        if isinstance(o, self.__class__):
            return self.vars != o.vars
        return NotImplemented

    def __forward_op(op):

        @wraps(op)
        def method(self: Self, *a):
            ident = self.vars
            if not a:
                return op(ident)
            elif isinstance(a := a[0], ProPredicate):
                return op(ident, a.vars)
            return NotImplemented

        return method

    __ge__, __gt__ = __forward_op(operator.ge), __forward_op(operator.gt)
    __le__, __lt__ = __forward_op(operator.le), __forward_op(operator.lt)

    __hash__ = __forward_op(hash)
    
    del __forward_op




class ProPredicate(_PredicateBase, _PredicateOpsMixin, _PredicateCompareMixin, ABC):
    
    __slots__ = ()
    
    vars: tuple[_T_PredVar]






@ProPredicate.register
class ProEnumPredicate(_PredicateBase, _PredicateOpsMixin, _PredicateCompareMixin):

    __slots__ = ()
    __setattr__ = object.__setattr__

    @property
    def _rawvalue_(self):
        return self.vars[0]

    # def __copy__(self):
    #     return self



class AccessLevel(ProEnumPredicate, Enum):  
    """Access level for dependencies

    Attributes:
        public (AccessLevel): public
        protected (AccessLevel): protected
        guarded (AccessLevel): guarded
        private (AccessLevel): private
    """

    public: "AccessLevel" = 1
    protected: "AccessLevel" = 2
    guarded: "AccessLevel" = 3
    private: "AccessLevel" = 4
    

    @classmethod
    def _missing_(cls, val):
        if val in _empty_access_levels:
            return cls.public
        elif val in _access_lavel_rawvalues:
            return _access_lavel_rawvalues[val] 
        return super()._missing_(val)

    def pro_entries(self, it: abc.Iterable['Container'], scope: 'DepGraph', src: 'DepSrc') -> abc.Iterable['Container']:
        return tuple(c for c in it if self in c.access_level(src.container))

    def __contains__(self, obj) -> bool:
        return isinstance(obj, AccessLevel) and self.vars >= obj.vars



_empty_access_levels = frozenset((None, 0, (0,), (None,)))
_access_lavel_rawvalues = {l._rawvalue_: l for l in AccessLevel}


class ScopePredicate(ProEnumPredicate, Enum):
    """The context in which to provider resolution.
    """

    only_self: "ScopePredicate" = True
    """Only inject from the current scope without considering parents"""

    skip_self: "ScopePredicate" = False
    """Skip the current scope and resolve from it's parent instead."""
    
    @classmethod
    def _missing_(cls, val):
        if val in _scope_predicate_rawvalues:
            return _scope_predicate_rawvalues[val] 
        return super()._missing_(val)

    def pro_entries(self, it: abc.Iterable['Container'], scope: 'DepGraph', src: 'DepSrc') -> abc.Iterable['Container']:
        return it if (scope is src.graph) is self._rawvalue_ else ()


_scope_predicate_rawvalues = {l._rawvalue_: l for l in ScopePredicate}



PUBLIC: AccessLevel = AccessLevel.public
"""public access level"""

PROTECTED: AccessLevel = AccessLevel.protected
"""protected access level"""

GUARDED: AccessLevel = AccessLevel.guarded
"""guarded access level"""

PRIVATE: AccessLevel = AccessLevel.private
"""private access level"""

ONLY_SELF: ScopePredicate = ScopePredicate.only_self
"""Only inject from the current scope without considering parents"""

SKIP_SELF: ScopePredicate = ScopePredicate.skip_self
"""Skip the current scope and resolve from it's parent instead."""



@private_setattr
class ProNoopPredicate(ProPredicate):

    __slots__ = ()

    vars = ()
    __pred = None

    def __new__(cls):
        if self := cls.__pred:
            return self
        self = cls.__pred = _object_new(cls)
        return self

    def pro_entries(self, it: abc.Iterable['Container'], *args) -> abc.Iterable['Container']:
        return it



@private_setattr
class ProOperatorPredicate(ProPredicate):

    __slots__ = ()

    vars: tuple[ProPredicate]

    @staticmethod
    @abstractmethod
    def operate(pros: abc.Set['Container']) -> abc.Iterable['Container']: ...

    def _reduce(self, it: abc.Iterable[abc.Iterable['Container']]):
        return reduce(self.operate, it)

    def pro_entries(self, it: abc.Iterable['Container'], *args) -> abc.Iterable['Container']:
        it = tuple(it)
        res = self._reduce({*pred.pro_entries(it, *args)} for pred in self.vars)
        return tuple(sorted(res, key=it.index))




class ProOrPredicate(ProOperatorPredicate):

    __slots__ = ()

    operate = staticmethod(operator.or_)
   
   

class ProAndPredicate(ProOperatorPredicate):
    
    __slots__ = ()

    operate = staticmethod(operator.and_)
   

class ProSubPredicate(ProOperatorPredicate):

    __slots__ = ()

    operate = staticmethod(operator.sub)



class ProInvertPredicate(ProSubPredicate):

    __slots__ = ()

    def __new__(cls: type[Self], *right: _T_Pred) -> Self:
        return super().__new__(cls, _noop_pred, *right)

    def __copy__(self):
        return self.__class__(*self.vars[1:])

    def __reduce__(self):
        return self.__class__, self.vars[1:]






class ProSlice(ProPredicate, t.Generic[_T_Start, _T_Stop, _T_Step]):
    """Represents a slice or the _Provider resolution order_"""

    __slots__ = ()

    vars: tuple[_T_Start, _T_Stop, _T_Step]

    def __new__(cls: type[Self], start: _T_Start=None, stop: _T_Stop=None, step: _T_Step=None) -> Self:
        return super().__new__(cls, start, stop, step)

    @property
    def start(self):
        return self.vars[0]

    @property
    def stop(self):
        return self.vars[1]

    @property
    def step(self):
        return self.vars[2]

    def pro_entries(self, it: abc.Iterable['Container'], scope: 'DepGraph', src: 'DepSrc') -> abc.Iterable['Container']:
        it = tuple(it)
        start, stop, step = self.vars
        if isinstance(start, ProPredicate):
            start = it.index(next(iter(start.pro_entries(it, scope, src)), None))
        
        if isinstance(stop, ProPredicate):
            stop = it.index(next(iter(stop.pro_entries(it, scope, src)), None))

        return it[start:stop:step]        

    def __repr__(self) -> str:
        start, stop, step = self.start, self.stop, self.step
        return f'[{start}:{stop}:{step}]'




_T_FilterPred = t.TypeVar('_T_FilterPred', bound=abc.Callable[..., bool], covariant=True)
class ProFilter(ProPredicate):
    
    __slots__ = ()

    def __new__(cls: type[Self], filter: _T_FilterPred, extra_args: int=None) -> Self:
        if extra_args is None:
            try:
                sig = signature(filter)    
            except Exception: 
                if not callable(filter):
                    raise
                extra_args = 0 # pragma: no-cover
            else:
                extra_args = len(sig.parameters) - 1
                if extra_args > 1 or any(p.kind is Parameter.VAR_POSITIONAL for p in sig.parameters.values()):
                    extra_args = max(2, extra_args)
        return super().__new__(cls, filter, extra_args)

    def pro_entries(self, it: abc.Iterable['Container'], *args) -> abc.Iterable['Container']:
        fn, ln = self.vars
        args = args[:ln]
        return tuple(c for c in it if fn(c, *args))



_noop_pred = ProNoopPredicate()





@private_setattr
class PureDep(DependencyMarker, t.Generic[T_Injectable]):
    """Explicitly marks given injectable as a dependency. 

    Attributes:
        abstract (T_Injectable): the marked dependency.

    Params:
        abstract (T_Injectable): the dependency to mark.
    """
    
    __slots__ = "_ident",

    _ident: T_Injected

    predicate: ProPredicate = _noop_pred
    default: T_Default = Missing

    has_default: bool = False
    injects_default: bool = False

    def __new__(cls: type[Self], abstract: T_Injectable) -> Self:
        if abstract.__class__ is cls:
            return abstract
        self = _object_new(cls)
        self.__setattr(_ident=abstract)
        return self

    @property
    def abstract(self) -> T_Injectable:
        return self._ident

    @property
    def lookup(self):
        return Lookup(self)

    @property
    def __origin__(self): return self.__class__

    def __copy__(self):
        return self

    def __reduce__(self):
        return self.__class__, (self._ident,)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.abstract!s})"

    def __init_subclass__(cls, *args, **kwargs):
        if not cls.__module__.startswith(__package__): # pragma: no cover
            raise TypeError(f"Cannot subclass {cls.__module__}.{cls.__qualname__}")
    
    def _as_dict(self):
        return {
            'abstract': self.abstract,
            'predicate': self.predicate,
            'default': self.default,
        }

    def replace(self, **kwds):
        return Dep(*(self._as_dict() | kwds).values())

    def __eq__(self, x) -> bool:
        cls = self.__class__
        if cls is PureDep:
            return self._ident == x
        elif isinstance(x, cls):
            return self._ident == x._ident
        return NotImplemented

    def __ne__(self, x) -> bool:
        cls = self.__class__
        if cls is PureDep:
            return self._ident != x
        elif isinstance(x, cls):
            return self._ident != x._ident
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._ident)

    def __and__(self, x) -> Self:
        if isinstance(x, ProPredicate):
            return self.replace(predicate=self.predicate & x)
        return NotImplemented
    
    def __rand__(self, x) -> Self:
        if isinstance(x, ProPredicate):
            return self.replace(predicate=x & self.predicate)
        return NotImplemented
    
    def __or__(self, x) -> Self:
        if isinstance(x, ProPredicate):
            return self.replace(predicate=self.predicate | x)
        return NotImplemented

    def __ror__(self, x) -> Self:
        if isinstance(x, ProPredicate):
            return self.replace(predicate=x | self.predicate)
        return NotImplemented

    def __invert__(self) -> Self:
        return self.replace(predicate=ProInvertPredicate(self.predicate))
        




_pure_dep_default_set = frozenset([
    (PureDep.predicate, PureDep.default),
])

@private_setattr
class Dep(PureDep):

    """Marks an injectable as a `dependency` to be injected."""

    __slots__ = '_ash'

    def __new__(
        cls: type[Self],
        abstract: T_Injectable,
        predicate: ProPredicate = ProNoopPredicate(),
        default = Missing,
    ):
        
        ident = abstract, predicate or _noop_pred, default
        if ident[1:] in _pure_dep_default_set:
            if abstract.__class__ in (cls, PureDep):
                return abstract
            return PureDep(abstract)

        self = _object_new(cls)
        self.__setattr(_ident=ident)
        return self

    @property
    def __origin__(self):
        return self.__class__

    @property
    def abstract(self):
        return self._ident[0]

    @property
    def predicate(self):
        return self._ident[1]

    @property
    def default(self):
        return self._ident[2]

    @property
    def has_default(self):
        return not self.default is Missing

    @property
    def injects_default(self):
        return isinstance(self.default, DependencyMarker)

    def __reduce__(self):
        return self.__class__, self._ident

    def __hash__(self):
        try:
            return self._ash
        except AttributeError:
            self.__setattr(_ash=hash(self._ident))
            return self._ash

    def __repr__(self) -> str:
        abstract, predicate, default = self.abstract, self.predicate, self.default
        return f'{self.__class__.__qualname__}({abstract=}, {predicate=!r}, {default=!r})'



class Lookup(DependencyMarker, BaseLookup):
    """Represents a lazy lookup of a given dependency.

    Attributes:
        __abstract__ (Injectable): the dependency to lookup.

    Params:
        abstract (Injectable): the dependency to lookup.
    """

    __slots__ = ()
    __offset__ = 1

    @t.overload
    def __new__(cls: type[Self], abstract: type[T_Injected]) -> Self: ...

    __new__ = BaseLookup.__new__

    @property
    def __abstract__(self) -> type[T_Injected]:
        return self.__expr__[0]

    @property
    def __origin__(self): return self.__class__





import logging
import sys
import typing as t
from collections.abc import Callable
from types import MethodType

from typing_extensions import Self


from . import providers
from .markers import Injectable, T_Injectable, T_Injected
from .exceptions import InjectorLookupError
from ._common import Missing, FrozenDict, ReadonlyDict, private_setattr
from ._bindings import Binding


if t.TYPE_CHECKING: # pragma: no cover
    from .graph import DepGraph, NullGraph
    from .scopes import Scope


logger = logging.getLogger(__name__)

_T = t.TypeVar('_T')
_object_new = object.__new__


TContextBinding = Callable[
    ["Injector", t.Optional[Injectable]], Callable[..., T_Injected]
]



@private_setattr
class Injector(ReadonlyDict[T_Injectable, Callable[[], T_Injected]]):
    """An isolated dependency injection context for a given `Scope`. 
    
    Attributes:
        graph (DepGraph): the dependency graph for this injector
        parent (Injector): a parent injector to provide missing dependencies.

    Params:
        graph (DepGraph): the dependency graph for this injector
        parent (Injector): a parent injector to provide missing dependencies.

    """
    __slots__ = 'graph', 'parent', '__weakref__',

    graph: "DepGraph"
    parent: Self

    def __init__(self, graph: 'DepGraph', parent: Self):
        self.__setattr(graph=graph, parent= parent)

    @property
    def name(self) -> str:
        """The name of the scope. Usually returns the injector's `scope.name` 
        """
        return self.graph.name

    def bound(self, abstract: T_Injectable) -> T_Injected:
        return self[self.graph[abstract]]

    def make(self, abstract: T_Injectable, /, *args, **kwds) -> T_Injected:
        graph = self.graph
        if dep := graph[abstract]:
            return self[dep](*args, **kwds)
        elif callable(abstract):
            if prov := getattr(abstract, '__uzi_provider__', None):
                if dep := graph[prov]:
                    return self[dep](*args, **kwds)
            else:
                prov = providers.Partial(abstract)
                setattr(abstract, '__uzi_provider__', prov)
                return self[graph[prov]](*args, **kwds)
        else:
            return self[abstract](*args, **kwds)

    def __bool__(self):
        return not not self.graph
        
    def __contains__(self, x) -> bool:
        return self.__contains(x) or x in self.parent

    def __missing__(self, dep: Binding):
        try:
            return self.__setdefault(dep, (dep.graph is self.graph and dep.bind(self)) or self.parent[dep])   
        except AttributeError as e:
            raise TypeError(f'Injector key must be a `Dependency` not `{dep.__class__.__qualname__}`')
        
    __setdefault = dict.setdefault
    __contains = dict.__contains__

    def close(self): ...            

    def copy(self):
        return self

    __copy__ = copy

    def __reduce__(self): 
        raise TypeError(f'cannot copy `{self.__class__.__name__}`')

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name!r})"

    def __repr__(self) -> str:
        return f"<{self!s}, {self.parent!r}>"
    
    def __hash__(self):
        return id(self)

    def __eq__(self, x):
        return x is self if isinstance(x, Injector) else NotImplemented

    def __ne__(self, x):
        return not x is self if isinstance(x, Injector) else NotImplemented







class NullInjector(Injector):
    """A 'noop' `Injector` used as the parent of root injectors.  

    Attributes:
        scope (NullScope): the scope 
        parent (None): The parent injector

    Params:
        None
    """

    __slots__ = ()

    parent: t.Final = None
    _scope: 'NullGraph' = None

    @property
    def scope(self):
        if not (scp := self._scope) is None:
            return scp
        else:
            from uzi.scopes import _null_scope
            scp = self.__class__._scope = _null_scope
            return scp

    @property
    def graph(self):
        return self.scope.graph

    def __init__(self, *a, **kw) -> None: ...
    
    def __reduce__(self): return self.__class__, ()

    def __getitem__(self, dep: Binding):
        try:
            dep.bind(self)
        except (AttributeError, TypeError) as e:
            raise InjectorLookupError(dep) from e
        else:
            raise InjectorLookupError(dep)
    
    def __bool__(self, *a, **kw): 
        return False

    __contains__ = __bool__
    
    def __eq__(self, o) -> bool:
        return o.__class__ is self.__class__ 

    def __ne__(self, o) -> bool:
        return not o.__class__ is self.__class__ 

    __hash__ = classmethod(hash)
   
   
    

_null_injector = NullInjector()


_T = t.TypeVar("_T")
_T_Fn = t.TypeVar("_T_Fn", bound=Callable)



class _InjectorExitStack(list[tuple[bool, _T_Fn]]): # pragma: no cover
    """Async context manager for dynamic management of a stack of exit
    callbacks.

    For example:
        async with AsyncExitStack() as stack:
            connections = [await stack.enter_async_context(get_connection())
                for i in range(5)]
            # All opened connections will automatically be released at the
            # end of the async with statement, even if attempts to open a
            # connection later in the list raise an exception.
    """
    __slots__ = ()

    @staticmethod
    def _create_exit_wrapper(cm, cm_exit):
        return MethodType(cm_exit, cm)

    @staticmethod
    def _create_cb_wrapper(callback, /, *args, **kwds):
        def _exit_wrapper(exc_type, exc, tb):
            callback(*args, **kwds)
        return _exit_wrapper

    @staticmethod
    def _create_async_exit_wrapper(cm, cm_exit):
        return MethodType(cm_exit, cm)

    @staticmethod
    def _create_async_cb_wrapper(callback, /, *args, **kwds):
        async def _exit_wrapper(exc_type, exc, tb):
            await callback(*args, **kwds)

        return _exit_wrapper

    def _push_exit_callback(self, k, cb, is_sync=True):
        self.append((is_sync, cb))

    def push(self, exit: _T) -> _T:
        """Registers a callback with the standard __exit__ method signature.

        Can suppress exceptions the same way __exit__ method can.
        Also accepts any object with an __exit__ method (registering a call
        to the method instead of the object itself).
        """
        # We use an unbound method rather than a bound method to follow
        # the standard lookup behaviour for special methods.
        _cb_type = exit.__class__

        try:
            exit_method = _cb_type.__exit__
        except AttributeError:
            # Not a context manager, so assume it's a callable.
            self._push_exit_callback(exit, exit)
        else:
            self._push_exit_callback(exit, self._create_exit_wrapper(exit, exit_method))
        return exit  # Allow use as a decorator.

    def enter(self, cm):
        """Enters the supplied context manager.

        If successful, also pushes its __exit__ method as a callback and
        returns the result of the __enter__ method.
        """
        # We look up the special methods on the type to match the with
        # statement.
        _cm_type = cm.__class__
        _exit = _cm_type.__exit__
        result = _cm_type.__enter__(cm)
        self._push_exit_callback(cm, self._create_exit_wrapper(cm, _exit))
        return result

    def callback(self, callback: _T_Fn, /, *args, **kwds) -> _T_Fn:
        """Registers an arbitrary callback and arguments.

        Cannot suppress exceptions.
        """
        _exit_wrapper = self._create_cb_wrapper(callback, *args, **kwds)

        # We changed the signature, so using @wraps is not appropriate, but
        # setting __wrapped__ may still help with introspection.
        _exit_wrapper.__wrapped__ = callback
        self._push_exit_callback((callback, args, kwds), _exit_wrapper)
        return callback  # Allow use as a decorator

    def close(self):
        """Immediately unwind the context stack."""
        self.__exit__(None, None, None)

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        received_exc = exc_details[0] is not None

        # We manipulate the exception state so it behaves as though
        # we were actually nesting multiple with statements

        # Callbacks are invoked in LIFO order to match the behaviour of
        # nested context managers
        suppressed_exc = pending_raise = False

        if self:
            frame_exc = sys.exc_info()[1]

            while self:
                is_sync, cb = self.pop()
                assert is_sync
                try:
                    if cb(*exc_details):
                        suppressed_exc = True
                        pending_raise = False
                        exc_details = (None, None, None)
                except:
                    new_exc_details = sys.exc_info()
                    # simulate the stack of exceptions by setting the context
                    _fix_exception_context(
                        new_exc_details[1], exc_details[1], frame_exc
                    )
                    pending_raise = True
                    exc_details = new_exc_details

            if pending_raise:
                try:
                    # bare "raise exc_details[1]" replaces our carefully
                    # set-up context
                    fixed_ctx = exc_details[1].__context__
                    raise exc_details[1]
                except BaseException:
                    exc_details[1].__context__ = fixed_ctx
                    raise
        return received_exc and suppressed_exc

    async def enter_async(self, cm):
        """Enters the supplied async context manager.

        If successful, also pushes its __aexit__ method as a callback and
        returns the result of the __aenter__ method.
        """
        _cm_type = cm.__class__
        _exit = _cm_type.__aexit__
        result = await _cm_type.__aenter__(cm)
        self._push_async_cm_exit(cm, _exit)
        return result

    def push_async_exit(self, exit):
        """Registers a coroutine function with the standard __aexit__ method
        signature.

        Can suppress exceptions the same way __aexit__ method can.
        Also accepts any object with an __aexit__ method (registering a call
        to the method instead of the object itself).
        """
        _cb_type = type(exit)
        try:
            exit_method = _cb_type.__aexit__
        except AttributeError:
            # Not an async context manager, so assume it's a coroutine function
            self._push_exit_callback(exit, exit, False)
        else:
            self._push_async_cm_exit(exit, exit_method)
        return exit  # Allow use as a decorator

    def callback_async(self, callback, /, *args, **kwds):
        """Registers an arbitrary coroutine function and arguments.

        Cannot suppress exceptions.
        """
        _exit_wrapper = self._create_async_cb_wrapper(callback, *args, **kwds)

        # We changed the signature, so using @wraps is not appropriate, but
        # setting __wrapped__ may still help with introspection.
        _exit_wrapper.__wrapped__ = callback
        self._push_exit_callback(callback, _exit_wrapper, False)
        return callback  # Allow use as a decorator

    async def aclose(self):
        """Immediately unwind the context stack."""
        await self.__aexit__(None, None, None)

    def _push_async_cm_exit(self, cm, cm_exit):
        """Helper to correctly register coroutine function to __aexit__
        method."""
        _exit_wrapper = self._create_async_exit_wrapper(cm, cm_exit)
        self._push_exit_callback(cm, _exit_wrapper, False)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc_details):
        # exc_details = = (None, None, None)
        received_exc = exc_details[0] is not None

        # We manipulate the exception state so it behaves as though
        # we were actually nesting multiple with statements

        # Callbacks are invoked in LIFO order to match the behaviour of
        # nested context managers
        suppressed_exc = pending_raise = False

        if self:
            frame_exc = sys.exc_info()[1]

            while self:
                is_sync, cb = self.pop()
                try:
                    if is_sync:
                        cb_suppress = cb(*exc_details)
                    else:
                        cb_suppress = await cb(*exc_details)

                    if cb_suppress:
                        suppressed_exc = True
                        pending_raise = False
                        exc_details = (None, None, None)
                except:
                    new_exc_details = sys.exc_info()
                    # simulate the stack of exceptions by setting the context
                    _fix_exception_context(
                        new_exc_details[1], exc_details[1], frame_exc
                    )
                    pending_raise = True
                    exc_details = new_exc_details

            if pending_raise:
                try:
                    # bare "raise exc_details[1]" replaces our carefully
                    # set-up context
                    fixed_ctx = exc_details[1].__context__
                    raise exc_details[1]
                except BaseException:
                    exc_details[1].__context__ = fixed_ctx
                    raise
        return received_exc and suppressed_exc




def _fix_exception_context(new_exc, old_exc, frame_exc): # pragma: no cover
    # Context may not be correct, so find the end of the chain
    while 1:
        exc_context = new_exc.__context__
        if exc_context is None or exc_context is old_exc:
            # Context is already set correctly (see issue 20317)
            return
        if exc_context is frame_exc:
            break
        new_exc = exc_context
    # Change the end of the chain to point to the exception
    # we expect it to reference
    new_exc.__context__ = old_exc



from blinker import NamedSignal


on_container_create = NamedSignal(f'{__package__}.on_container_create')
on_container_init = NamedSignal(f'{__package__}.on_container_init')
on_provider_registered = NamedSignal(f'{__package__}.on_provider_registered')
on_injector_init = NamedSignal(f'{__package__}.on_injector_init')
on_scope_init = NamedSignal(f'{__package__}.on_scope_init')




import attr


class UziException(Exception):
    """Base class for all internal exceptions."""




@attr.s()
class FinalProviderOverrideError(TypeError, UziException):
    """Raised when a final provider has an override.
    """
    
    abstract: 'Injectable' = attr.ib(default=None)
    final: 'Provider' = attr.ib(default=None)
    overrides: tuple['Provider' ]= attr.ib(default=(), converter=tuple)


class InvalidStateError(UziException):
    """The operation is not allowed in this state."""



@attr.s()
class InjectorLookupError(KeyError, UziException):
    """Raised by ~Injector` when a missing dependency is requested.
    
    Args:
        abstract (Injectable): the missing dependency
    """

    abstract: "Injectable" = attr.ib(default=None)
    scope: "DepGraph" = attr.ib(default=None)





class ProError(TypeError, UziException):
    """Raised when there is an issue with provider resolution order (`pro`) 
        consistency
    """





from .providers import Provider
from .markers import Injectable
from .graph import DepGraph



from . import signals


from .containers import Container  
from .providers import AnnotationProvider, UnionProvider, DepMarkerProvider, LookupMarkerProvider  





@signals.on_container_create.connect
def _register_implicit_providers(sender: type[Container], container: Container):
    provs = UnionProvider(), DepMarkerProvider(), AnnotationProvider(), LookupMarkerProvider()
    for prov in provs:
        container[prov.abstract] = prov

